@file:OptIn(ExperimentalContracts::class)

package ru.landrafhomyak.utility.reference_counter

import kotlin.contracts.ExperimentalContracts
import kotlin.contracts.InvocationKind
import kotlin.contracts.contract
import kotlinx.atomicfu.AtomicLong
import kotlinx.atomicfu.atomic
import kotlinx.atomicfu.getAndUpdate
import kotlinx.atomicfu.update
import ru.landgrafhomyak.utility.highlevel_try_finally.safeAutoClose1
import ru.landgrafhomyak.utility.highlevel_try_finally.safeAutoClose2

@CloseableReferenceCounter_Debug.RequiresExplicitDebug
public class CloseableReferenceCounter_Debug(
	private val _dbgName: String,
	private val _errMessage: String,
	private val _logger: Observer? = null,
) {
	public fun interface Observer {
		public fun observeState(instance: CloseableReferenceCounter_Debug, actions: String)
	}

	@RequiresOptIn(level = RequiresOptIn.Level.WARNING)
	@Retention(AnnotationRetention.BINARY)
	public annotation class RequiresExplicitDebug


	private val _value: AtomicLong = atomic(0L)

	@Suppress("RemoveRedundantQualifierName")
	private val _id = CloseableReferenceCounter_Debug._nextId.getAndUpdate(ULong::inc)

	@Suppress("NOTHING_TO_INLINE")
	private inline fun _throwErrors(valueToCheck: Long) {
		when {
			valueToCheck >= 0 || valueToCheck == _CloseableReferenceCounter_LowLevel.CLOSED_STATE_VALUE -> {}
			valueToCheck < _CloseableReferenceCounter_LowLevel.CLOSED_STATE_VALUE -> throw RuntimeException("Too many references")
			valueToCheck > _CloseableReferenceCounter_LowLevel.CLOSED_STATE_VALUE -> throw RuntimeException(".decref called more times than .incref")
		}
	}

	@RequiresExplicitDebug
	public fun throwErrors() {
		this._throwErrors(this._value.value)
	}

	public fun throwClosed() {
		throw IllegalStateException(this._errMessage)
	}

	public fun incref() {
		this._value.update { o ->
			if (o < 0) {
				this._throwErrors(o)
				this.throwClosed()
				this._throwErrors(o + 1)
			}
			return@update o + 1
		}
		this._logger?.observeState(this, "incref")
	}

	public inline fun <R> tryIncref(block: () -> R): R {
		contract {
			callsInPlace(block, InvocationKind.EXACTLY_ONCE)
		}
		this.incref()
		return safeAutoClose2(onError = this::decref, action = block)
	}

	public fun assertNotClosed() {
		if (this._value.value < 0) this.throwClosed()
	}

	public fun decref() {
		this._value.update { o ->
			if (o < 0) {
				this._throwErrors(o)
				this.throwClosed()
				this._throwErrors(o - 1)
			}
			return@update o - 1
		}
		this._logger?.observeState(this, "decref")
	}

	public inline fun <R> tryDecref(block: () -> R): R {
		contract {
			callsInPlace(block, InvocationKind.EXACTLY_ONCE)
		}
		this.assertNotClosed()
		return safeAutoClose2(onSuccess = this::decref, action = block)
	}

	public fun close(errExistRefs: String) {
		val state = _CloseableReferenceCounter_LowLevel.compareAndExchange(this._value, 0, _CloseableReferenceCounter_LowLevel.CLOSED_STATE_VALUE)
		this._throwErrors(state)
		when {
			state > 0 -> throw IllegalStateException(errExistRefs)
			state < 0 -> this.throwClosed()
		}
		this._logger?.observeState(this, "closed")
	}

	public inline fun <R> withRef(protected: () -> R): R {
		this.incref()
		return safeAutoClose1(finally = this::decref, action = protected)
	}

	override fun toString(): String {
		val refcntCached = this._value.value
		val stateRepr: String
		@Suppress("LiftReturnOrAssignment")
		when {
			refcntCached >= 0 -> stateRepr = refcntCached.toString()
			refcntCached == _CloseableReferenceCounter_LowLevel.CLOSED_STATE_VALUE -> stateRepr = "closed"
			refcntCached < _CloseableReferenceCounter_LowLevel.CLOSED_STATE_VALUE -> stateRepr = "overflow"
			refcntCached > _CloseableReferenceCounter_LowLevel.CLOSED_STATE_VALUE -> stateRepr = "underflow"
			else -> throw Error("Unreachable")
		}
		return "<ref counter \"${this._dbgName}@${this._id}\" [${stateRepr}]>"
	}

	public companion object {
		private val _nextId = atomic(0uL)
	}

	public object ObserveToStdout : Observer {
		override fun observeState(instance: CloseableReferenceCounter_Debug, actions: String) {
			print("${instance} ${actions}\n")
		}
	}
}